import torch
from torch import nn
import torch.nn.functional as F
from functools import partial
from torch.autograd import Variable


class MMDLoss(nn.Module):
    """
    Temp MMD loss implementation. Use the implementation from D-MMD.
    """
    def __init__(self):
        super().__init__()

    def pairwise_distance(self, x, y):
        if not len(x.shape) == len(y.shape) == 2:
            raise ValueError('Both inputs should be matrices.')

        if x.shape[1] != y.shape[1]:
            raise ValueError('The number of features should be the same.')

        x = x.view(x.shape[0], x.shape[1], 1)
        y = torch.transpose(y, 0, 1)
        output = torch.sum((x - y) ** 2, 1)
        output = torch.transpose(output, 0, 1)
        return output

    def gaussian_kernel_matrix(self, x, y, sigmas):
        sigmas = sigmas.view(sigmas.shape[0], 1)
        beta = 1. / (2. * sigmas)
        dist = self.pairwise_distance(x, y).contiguous()
        dist_ = dist.view(1, -1)
        s = torch.matmul(beta, dist_.cuda())
        return torch.sum(torch.exp(-s), 0).view_as(dist)

    def maximum_mean_discrepancy(self, x, y, kernel=gaussian_kernel_matrix):
        cost = torch.mean(kernel(x, x))
        cost += torch.mean(kernel(y, y))
        cost -= 2 * torch.mean(kernel(x, y))
        return cost

    def forward(self, s_feature, t_feature):
        sigmas = [
            1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 1, 5, 10, 15, 20, 25, 30, 35, 100,
            1e3, 1e4, 1e5, 1e6
        ]
        gaussian_kernel = partial(
            self.gaussian_kernel_matrix, sigmas=Variable(torch.cuda.FloatTensor(sigmas))  # TODO ?
        )
        loss_value = self.maximum_mean_discrepancy(s_feature, t_feature, kernel=gaussian_kernel)
        return loss_value


class WeightedMMDLoss(nn.Module):
    """
    Temp MMD loss implementation. Use the implementation from D-MMD.
    """
    def __init__(self):
        super().__init__()

    def pairwise_distance(self, x, y):
        if not len(x.shape) == len(y.shape) == 2:
            raise ValueError('Both inputs should be matrices.')

        if x.shape[1] != y.shape[1]:
            raise ValueError('The number of features should be the same.')

        x = x.view(x.shape[0], x.shape[1], 1)
        y = torch.transpose(y, 0, 1)
        output = torch.sum((x - y) ** 2, 1)
        output = torch.transpose(output, 0, 1)
        return output

    def gaussian_kernel_matrix(self, x, y, sigmas):
        sigmas = sigmas.view(sigmas.shape[0], 1)
        beta = 1. / (2. * sigmas)
        dist = self.pairwise_distance(x, y).contiguous()
        dist_ = dist.view(1, -1)
        s = torch.matmul(beta, dist_.cuda())
        return torch.sum(torch.exp(-s), 0).view_as(dist)

    def maximum_mean_discrepancy(self, x, y, x_weights, kernel=gaussian_kernel_matrix):
        # TODO chenck x_weights.shape[0]
        k1 = kernel(x, x)
        w1 = torch.matmul(x_weights.view(x_weights.shape[0], 1), x_weights.view(1, x_weights.shape[0]))
        kw1 = k1 * w1
        k2 = kernel(y, y)
        k3 = kernel(x, y)
        w3 = torch.matmul(x_weights.view(x_weights.shape[0], 1), torch.ones((1, x_weights.shape[0])).cuda())
        kw3 = k3 * w3
        cost = torch.mean(kw1)
        cost += torch.mean(k2)
        cost -= 2 * torch.mean(kw3)
        return cost

    def forward(self, s_feature, t_feature, s_weights):
        sigmas = [
            1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 1, 5, 10, 15, 20, 25, 30, 35, 100,
            1e3, 1e4, 1e5, 1e6
        ]
        gaussian_kernel = partial(
            self.gaussian_kernel_matrix, sigmas=Variable(torch.cuda.FloatTensor(sigmas))  # TODO ?
        )
        loss_value = self.maximum_mean_discrepancy(s_feature, t_feature, s_weights, kernel=gaussian_kernel)
        return loss_value
